#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#define MAXMEMORY 65536 /* maximum number of data words in memory */
#define NUMREGS 8 /* number of machine registers */
#define MAXLINELENGTH 1000

#define OP_SHIFT 13
#define A_SHIFT  10
#define B_SHIFT  7
#define C_SHIFT  0

#define ADD 0
#define ADDI 1
#define NAND 2
#define LUI 3
#define SW 4
#define LW 5
#define BNE 6
#define JALR 7

#define NOPINSTRUCTION ((short)0x0000)
#define HALTINSTRUCTION ((short)0xe071)

typedef struct IFIDStruct {
	short instr;
	short pc;
} IFID_t;

typedef struct IDEXStruct {
	short instr;
	short pc;
	short op0;
	short op1;
	short op2;
} IDEX_t;

typedef struct EXMEMStruct {
	short instr;
	short pc;
	short storeData;
	short aluResult;
	short branchTarget;
} EXMEM_t;

typedef struct MEMWBStruct {
	short instr;
	short pc;
	short writeData;
} MEMWB_t;

typedef struct WBENDStruct {
	short instr;
	short writeData;
} WBEND_t;

typedef struct stateStruct {
    short pc;
    short instrMem[MAXMEMORY];
    short dataMem[MAXMEMORY];
    short reg[NUMREGS];
    short numMemory;
    IFID_t IFID;
    IDEX_t IDEX;
    EXMEM_t EXMEM;
    MEMWB_t MEMWB;
    WBEND_t WBEND;
    short cycles; /* number of cycles run so far */
} state_t;

void printState(state_t *state);
short field_op(short instruction);
short field_ra(short instruction);
short field_rb(short instruction);
short field_rc(short instruction);
short field_imm7(short instruction);
short field_imm10(short instruction);
short target(short instr);
short source1(short instr);
short source2(short instr);
void printInstruction(short instr);
void print_rrrtype( short instr, char *name );
void print_rritype( short instr, char *name );
void print_ritype( short instr, char *name );
short signext(short num);
void run(state_t state);


main(int argc, char *argv[])
{
    short i, j;
    short statusNum;
    char line[MAXLINELENGTH];
    state_t state;
    FILE *filePtr;

    if (argc != 2) {
	printf("error: usage: %s <machine-code file>\n", argv[0]);
	exit(1);
    }

    memset(&state, 0, sizeof(state_t));

    /* read machine-code file into instruction/data memory (starting at address 0) */

    filePtr = fopen(argv[1], "r");
    if (filePtr == NULL) {
	printf("error: can't open file %s\n", argv[1]);
	perror("fopen");
	exit(1);
    }

    for (state.numMemory=0; fgets(line, MAXLINELENGTH, filePtr) != NULL;
	state.numMemory++) {
	if (sscanf(line, "%hx", &state.dataMem[state.numMemory]) != 1) {
	    printf("error in reading address %hd\n", state.numMemory);
	    exit(1);
	}
	state.instrMem[state.numMemory] = state.dataMem[state.numMemory];
	printf("memory[%hd]=%hx\n", state.numMemory, state.dataMem[state.numMemory]);
    }

    printf("%hd memory words\n", state.numMemory);

    printf("\tinstruction memory:\n");
    for (i=0; i<state.numMemory; i++) {
	printf("\t\tinstrMem[ %hd ] = ", i);
	printInstruction(state.instrMem[i]);
    }
 
    run(state);
}

void
printState(state_t *state)
{
    short i;
    printf("@@@\nstate before cycle %hd starts\n", state->cycles);
    printf("\tpc %hd\n", state->pc);

    printf("\tdata memory:\n");
	for (i=0; i<state->numMemory; i++) {
	    printf("\t\tdataMem[ %hd ] %hd\n", i, state->dataMem[i]);
	}
    printf("\tregisters:\n");
	for (i=0; i<NUMREGS; i++) {
	    printf("\t\treg[ %hd ] %hd\n", i, state->reg[i]);
	}
    printf("\tIFID:\n");
	printf("\t\tinstruction: "); printInstruction(state->IFID.instr);
	printf("\t\tpc %hd\n", state->IFID.pc);
    printf("\tIDEX:\n");
	printf("\t\tinstruction: "); printInstruction(state->IDEX.instr);
	printf("\t\tpc %d\n", state->IDEX.pc);
	printf("\t\toperand0 %d\n", state->IDEX.op0);
	printf("\t\toperand1 %d\n", state->IDEX.op1);
	printf("\t\toperand2 %d\n", state->IDEX.op2);
    printf("\tEXMEM:\n");
	printf("\t\tinstruction: "); printInstruction(state->EXMEM.instr);
	printf("\t\tpc %hd\n", state->EXMEM.pc);
	printf("\t\tstoreData %hd\n", state->EXMEM.storeData);
	printf("\t\taluResult %hd\n", state->EXMEM.aluResult);
	printf("\t\tbranchTarget %hd\n", state->EXMEM.branchTarget);
    printf("\tMEMWB:\n");
	printf("\t\tinstruction: "); printInstruction(state->MEMWB.instr);
	printf("\t\tpc %hd\n", state->MEMWB.pc);
	printf("\t\twriteData %hd\n", state->MEMWB.writeData);
    printf("\tWBEND:\n");
	printf("\t\tinstruction: "); printInstruction(state->WBEND.instr);
	printf("\t\twriteData %hd\n", state->WBEND.writeData);
}

short field_op(short instruction)
{
    return (instruction >> OP_SHIFT) & 0x7;
}

short
field_ra(short instruction)
{
    return (instruction >> A_SHIFT) & 0x7;
}

short
field_rb(short instruction)
{
    return (instruction >> B_SHIFT) & 0x7;
}

short
field_rc(short instruction)
{
    return (instruction >> C_SHIFT) & 0x7;
}

short
field_imm7(short instruction)
{
    return (instruction & 0x7F);
}

short
field_imm10(short instruction)
{
    return (instruction & 0x3FF);
}

void
printInstruction(short instr)
{
    char *opcodeString;

    if (instr == NOPINSTRUCTION) {
	printf("nop\n");
    } else if (instr == HALTINSTRUCTION) {
	printf("halt\n");
    } else if (field_op(instr) == ADD) {
	print_rrrtype(instr, "add");
    } else if (field_op(instr) == ADDI) {
	print_rritype(instr, "addi");
    } else if (field_op(instr) == NAND) {
	print_rrrtype(instr, "nand");
    } else if (field_op(instr) == LUI) {
	print_ritype(instr, "lui");
    } else if (field_op(instr) == SW) {
	print_rritype(instr, "sw");
    } else if (field_op(instr) == LW) {
	print_rritype(instr, "lw");
    } else if (field_op(instr) == BNE) {
	print_rritype(instr, "bne");
    } else if (field_op(instr) == JALR) {
	print_rritype(instr, "jalr");
    } else {
	/* note: this won't happen */
	printf("data: %hd\n", instr);
    }
}

void
print_rrrtype(instr, name)
short instr;
char *name;
{
    printf("%s %hd %hd %hd\n", name, field_ra(instr), field_rb(instr), field_rc(instr));
}

void
print_rritype(instr, name)
short instr;
char *name;
{
    printf("%s %hd %hd %hd\n", name, field_ra(instr), field_rb(instr), signext(field_imm7(instr)));
}

void
print_ritype(instr, name)
short instr;
char *name;
{
    printf("%s %hd %hd %hd\n", name, field_ra(instr), field_rb(instr), field_imm10(instr));
}


short
signext(short num)
{
    /* convert a 7 bit number into a 16-bit number */
    if (num & 0x40) {
	num -= 128;
    }
    return(num);
}

short
target(short instr)
{
    if (field_op(instr) == ADD || field_op(instr) == ADDI || field_op(instr) == LUI ||
	field_op(instr) == NAND || field_op(instr) == LW || field_op(instr) == JALR) {
	return field_ra(instr);
    } else {
	return 0;
    }
}

short
source1(short instr)
{
    if (field_op(instr) == ADD || field_op(instr) == ADDI || field_op(instr) == NAND ||
	field_op(instr) == BNE || field_op(instr) == SW || field_op(instr) == LW ||
	field_op(instr) == JALR) {
	return field_rb(instr);
    } else {
	return 0;
    }
}

short
source2(short instr)
{
    if (field_op(instr) == ADD || field_op(instr) == NAND) {
	return field_rc(instr);
    } else if (field_op(instr) == BNE || field_op(instr) == SW) {
	return field_ra(instr);
    } else {
	return 0;
    }
}

void
run(state_t state)
{
    state_t new;
    short MUXalu1out;
    short MUXalu2out;
    short i;

    memset(&new, 0, sizeof(state_t));

    while (1) {

	printState(&state);

	/* check for halt */
	if (state.MEMWB.instr == HALTINSTRUCTION) {
	    printf("machine halted\n");
	    printf("total of %hd cycles executed\n", state.cycles);

	    exit(0);
	}

	memcpy(&new, &state, sizeof(state_t)); /* copy everything so all we have to do is make changes.
					    (this is primarily for the memory and reg arrays) */
	new.cycles++;

	/* --------------------- IF stage --------------------- */

	//      in this section you can read from
	//          state.instrMem
	//          state.pc
	//	and can write to
	//          new.IFID
	//          new.pc

	CODE DELETED

	/* --------------------- ID stage --------------------- */

	//        in this section you can read from
	//            state.IFID
	//        and can write to
	//            new.IDEX
	//
	//        you will also need to insert a bubble (NOP) in the pipeline if
	//        you recognize that the instruction currently in the EX stage is
	//        a LW and the instruction currently in this stage depends on the
	//        result of that load.  therefore, you will be able to read from
	//        state.IDEX.instr as well to determine whether or not you need to 
	//        insert a bubble into the pipeline.

	/* insert bubble if the previous instruction is a LOAD and
	 * this instruction depends on the result of the LOAD */

	CODE DELETED

	/* --------------------- EX stage --------------------- */

	//        in this section you can read from
	//            state.IDEX
	//        and can write to
	//            new.EXMEM
	//
	//        you will handle forwarding here, so you will also need to read 
	//        from later pipeline stages: state.EXMEM, state.MEMWB, state.MEMWB.
	//        if the data that you need is from an instruction that is still in 
	//        the pipeline, you will need to obtain the operand from a later
	//        pipeline stage instead of the state.IDEX (i.e. the values just
	//        read out of the register file).
	//
	//        you will need to resolve branch prediction at this point, so you
	//        may need to alter things like state.pc as well as some earlier
	//        stages in the pipeline (state.IFID, state.IDEX) in the case that
	//        the static prediction was incorrect.  if we guessed wrong (e.g. we
	//        predict that the branch SHOULD be taken but it is NOT), there are
	//        instructions in the pipeline that should not be there.  in this
	//        case, we should squash them (turn them into NOPs).

	/* deal with forwarding:
	 * look ahead at the other instructions in the pipeline to see
	 * who (if anyone) is generating the operands we need.  put the
	 * pointers into the forwarding slots */

	CODE DELETED

	/* --------------------- MEM stage --------------------- */

	//        in this section you can read from
	//            state.EXMEM
	//            state.dataMem[] (data memory)
	//        and can write to
	//            new.MEMWB
	//            new.dataMem[] (data memory)

	CODE DELETED

	/* --------------------- WB stage --------------------- */

	//        in this section you can read from
	//            state.MEMWB
	//        and can write to
	//            new.WBEND
	//            new.reg[] (the register file)

	CODE DELETED

	/* --------------------- end stage --------------------- */

	/* transfer new. state into current state */
	memcpy(&state, &new, sizeof(state_t));

    }
}
